use crate::{db::DBConnection, DBResult};

use super::EntityDatum;

impl EntityDatum for String {
    fn sql_type() -> &'static str {
        "text"
    }

    fn bind_to<'a>(&self, stmt: &mut sqlite::Statement<'a>, index: usize) {
        stmt.bind((index, self.as_str()))
            .expect("couldn't bind string");
    }

    fn build_from<'a>(
        _: &DBConnection,
        _: &'static str,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        Ok((stmt.read(index)?, index + 1))
    }
}

impl EntityDatum for usize {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to<'a>(&self, stmt: &mut sqlite::Statement<'a>, index: usize) {
        stmt.bind((index, *self as i64))
            .expect("couldn't bind usize as i64");
    }

    fn build_from<'a>(
        _: &DBConnection,
        _: &'static str,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        Ok((stmt.read::<i64, _>(index)? as usize, index + 1))
    }
}

impl EntityDatum for isize {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, _index: usize) {
        todo!()
    }

    fn build_from<'a>(
        _: &DBConnection,
        _: &'static str,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        Ok((stmt.read::<i64, _>(index)? as isize, index + 1))
    }
}

impl EntityDatum for u64 {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, _index: usize) {
        todo!()
    }

    fn build_from<'a>(
        _: &DBConnection,
        _: &'static str,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        Ok((stmt.read::<i64, _>(index)? as u64, index + 1))
    }
}

impl EntityDatum for i64 {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, _index: usize) {
        todo!()
    }

    fn build_from<'a>(
        _: &DBConnection,
        _: &'static str,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        Ok((stmt.read(index)?, index + 1))
    }
}

impl<T: EntityDatum> EntityDatum for Option<T> {
    fn sql_type() -> &'static str {
        T::sql_type()
    }

    fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, _index: usize) {
        todo!()
    }

    fn build_from<'a>(
        _: &DBConnection,
        _: &'static str,
        _stmt: &mut sqlite::Statement<'a>,
        _index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        // Ok((stmt.read::<i64, _>(index)? as u64, index+1))
        todo!()
    }
}
