use super::{EntityDatum, EntityDatumList, EntityDatumVisitor};
use crate::{db::DBConnection, DBResult};

impl EntityDatumList for () {
    fn accept(&self, _: &mut impl EntityDatumVisitor) {}

    fn build_from<'a>(
        _conn: &DBConnection,
        _ctx: &'static str,
        _stmt: &mut sqlite::Statement<'a>,
    ) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(())
    }
}

impl<T: EntityDatum> EntityDatumList for T {
    fn accept(&self, visitor: &mut impl EntityDatumVisitor) {
        visitor.visit(self);
    }

    fn build_from<'a>(
        conn: &DBConnection,
        ctx: &'static str,
        stmt: &mut sqlite::Statement<'a>,
    ) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(T::build_from(conn, ctx, stmt, 0)?.0)
    }
}

impl<T0: EntityDatum> EntityDatumList for (T0,) {
    fn accept(&self, visitor: &mut impl EntityDatumVisitor) {
        visitor.visit(&self.0);
    }

    fn build_from<'a>(
        conn: &DBConnection,
        ctx: &'static str,
        stmt: &mut sqlite::Statement<'a>,
    ) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok((T0::build_from(conn, ctx, stmt, 0)?.0,))
    }
}

impl<T0: EntityDatum, T1: EntityDatum> EntityDatumList for (T0, T1) {
    fn accept(&self, visitor: &mut impl EntityDatumVisitor) {
        visitor.visit(&self.0);
        visitor.visit(&self.1);
    }

    fn build_from<'a>(
        conn: &DBConnection,
        ctx: &'static str,
        stmt: &mut sqlite::Statement<'a>,
    ) -> DBResult<Self>
    where
        Self: Sized,
    {
        let (d0, idx) = T0::build_from(conn, ctx, stmt, 0)?;
        let (d1, _) = T1::build_from(conn, ctx, stmt, idx)?;
        Ok((d0, d1))
    }
}

impl<T0: EntityDatum, T1: EntityDatum, T2: EntityDatum> EntityDatumList for (T0, T1, T2) {
    fn accept(&self, visitor: &mut impl EntityDatumVisitor) {
        visitor.visit(&self.0);
        visitor.visit(&self.1);
        visitor.visit(&self.2);
    }

    fn build_from<'a>(
        conn: &DBConnection,
        ctx: &'static str,
        stmt: &mut sqlite::Statement<'a>,
    ) -> DBResult<Self>
    where
        Self: Sized,
    {
        let (d0, idx) = T0::build_from(conn, ctx, stmt, 0)?;
        let (d1, idx) = T1::build_from(conn, ctx, stmt, idx)?;
        let (d2, _) = T2::build_from(conn, ctx, stmt, idx)?;
        Ok((d0, d1, d2))
    }
}

impl<T0: EntityDatum, T1: EntityDatum, T2: EntityDatum, T3: EntityDatum> EntityDatumList
    for (T0, T1, T2, T3)
{
    fn accept(&self, visitor: &mut impl EntityDatumVisitor) {
        visitor.visit(&self.0);
        visitor.visit(&self.1);
        visitor.visit(&self.2);
        visitor.visit(&self.3);
    }

    fn build_from<'a>(
        conn: &DBConnection,
        ctx: &'static str,
        stmt: &mut sqlite::Statement<'a>,
    ) -> DBResult<Self>
    where
        Self: Sized,
    {
        let (d0, idx) = T0::build_from(conn, ctx, stmt, 0)?;
        let (d1, idx) = T1::build_from(conn, ctx, stmt, idx)?;
        let (d2, idx) = T2::build_from(conn, ctx, stmt, idx)?;
        let (d3, _) = T3::build_from(conn, ctx, stmt, idx)?;
        Ok((d0, d1, d2, d3))
    }
}
