use super::{EntityPart, EntityPartList, EntityPartVisitor};

impl EntityPartList for () {
    type DatumList = ();
    fn accept_part_visitor(_: &mut impl EntityPartVisitor) {}
    fn accept_part_visitor_ref(_: &Self::DatumList, _: &mut impl EntityPartVisitor) {}
}

impl<P0: EntityPart> EntityPartList for P0 {
    type DatumList = P0::Datum;
    fn accept_part_visitor(v: &mut impl EntityPartVisitor) {
        v.visit::<P0>();
    }
    fn accept_part_visitor_ref(datum_list: &Self::DatumList, v: &mut impl EntityPartVisitor) {
        v.visit_datum::<P0>(datum_list);
    }
}

impl<P0: EntityPart> EntityPartList for (P0,) {
    type DatumList = P0::Datum;
    fn accept_part_visitor(v: &mut impl EntityPartVisitor) {
        v.visit::<P0>();
    }
    fn accept_part_visitor_ref(datum_list: &Self::DatumList, v: &mut impl EntityPartVisitor) {
        v.visit_datum::<P0>(datum_list);
    }
}

impl<P0: EntityPart, P1: EntityPart> EntityPartList for (P0, P1) {
    type DatumList = (P0::Datum, P1::Datum);
    fn accept_part_visitor(v: &mut impl EntityPartVisitor) {
        v.visit::<P0>();
        v.visit::<P1>();
    }
    fn accept_part_visitor_ref(datum_list: &Self::DatumList, v: &mut impl EntityPartVisitor) {
        v.visit_datum::<P0>(&datum_list.0);
        v.visit_datum::<P1>(&datum_list.1);
    }
}

impl<P0: EntityPart, P1: EntityPart, P2: EntityPart> EntityPartList for (P0, P1, P2) {
    type DatumList = (P0::Datum, P1::Datum, P2::Datum);
    fn accept_part_visitor(v: &mut impl EntityPartVisitor) {
        v.visit::<P0>();
        v.visit::<P1>();
        v.visit::<P2>();
    }
    fn accept_part_visitor_ref(datum_list: &Self::DatumList, v: &mut impl EntityPartVisitor) {
        v.visit_datum::<P0>(&datum_list.0);
        v.visit_datum::<P1>(&datum_list.1);
        v.visit_datum::<P2>(&datum_list.2);
    }
}

impl<P0: EntityPart, P1: EntityPart, P2: EntityPart, P3: EntityPart> EntityPartList
    for (P0, P1, P2, P3)
{
    type DatumList = (P0::Datum, P1::Datum, P2::Datum, P3::Datum);
    fn accept_part_visitor(v: &mut impl EntityPartVisitor) {
        v.visit::<P0>();
        v.visit::<P1>();
        v.visit::<P2>();
        v.visit::<P3>();
    }
    fn accept_part_visitor_ref(datum_list: &Self::DatumList, v: &mut impl EntityPartVisitor) {
        v.visit_datum::<P0>(&datum_list.0);
        v.visit_datum::<P1>(&datum_list.1);
        v.visit_datum::<P2>(&datum_list.2);
        v.visit_datum::<P3>(&datum_list.3);
    }
}
